; DPMS Screen Saver
;
; This program will work ONLY in DOS mode.
;
; It is capable of putting the monitor in two DPMS compliant Power Saving
; modes:
;
; 1> Suspend - Short image recovery time; Vertical Sync signal is off
; 2> Power Off - Longer image recovery time; Saves even more power than
;                suspend. Horizontal & Vertical Sync signals are off.
;

CSEG SEGMENT BYTE PUBLIC 'CODE'
     ASSUME CS:CSEG, DS:CSEG, ES:CSEG, SS:CSEG

     ORG 100h

ENTPT: jmp MAIN
;--------------------------------GENERAL EQUATES------------------------------
CR        equ  13
LF        equ  10
ENDSTR    equ  '$'

CMD_LEN   equ  80h
CMD_LINE  equ  81h

NORMAL    equ  0
POWER_OFF equ  1
SUSPEND   equ  2

CRTC_INDEX     equ  3D4h
CRTC_DATA      equ  3D5h

DEFAULT_SUSPEND_TIME     equ  60
DEFAULT_POWER_OFF_TIME   equ  300

;--------------------------------DATA (RESIDENT)------------------------------
CURRENT_MODE   db   NORMAL

SUSPEND_TIME   dw   DEFAULT_SUSPEND_TIME
POWER_OFF_TIME dw   DEFAULT_POWER_OFF_TIME

TIMER          dw   0

ORIG_08h_VECT  dd   ?
ORIG_09h_VECT  dd   ?

OLD_VERT_TOTAL      db   ?
OLD_HORIZ_TOTAL     db   ?
OLD_OVERFLOW        db   ?

OLD_SS         dw   ?     
OLD_SP         dw   ?
OLD_DX         dw   ?
OLD_AX         dw   ?

OLD_CRTC_IDX   db   ?
OLD_VSYNC      db   ?

;--------------------------------CODE (RESIDENT)------------------------------

INIT_WRITE PROC NEAR
     mov OLD_DX, dx                          ; Save the value of AX
     mov OLD_AX, ax                          ; and DX

     mov dx, CRTC_INDEX                      ; Save the old CRT Controller
     in al, dx                               ; index
     mov OLD_CRTC_IDX, al

     mov al, 11h                             ; Bit 7 of VSYNC register
     out dx, al                              ; is write protect. Disable
     mov dx, CRTC_DATA                       ; write protect
     in al, dx
     mov OLD_VSYNC, al
     and al, 01111111b
     out dx, al
     
     ret
INIT_WRITE ENDP

END_WRITE PROC NEAR
     mov dx, CRTC_INDEX                      
     mov al, 11h                             ; Restore write protect
     out dx, al                              ; to old value
     mov dx, CRTC_DATA
     mov al, OLD_VSYNC
     out dx, al

     mov dx, CRTC_INDEX                      ; Restore old value of CRTC
     mov al, OLD_CRTC_IDX                    ; index
     out dx, al
                                             ; Restore old AX and DX
     mov ax, OLD_AX
     mov dx, OLD_DX

     ret
END_WRITE ENDP

STOP_V_SYNC PROC NEAR
     call INIT_WRITE                         ; Save regs & clear Write protect

     mov dx, CRTC_INDEX
     mov al, 06h                             ; Suppress Vertical Sync signal
     out dx, al                              ; By setting vertical total
     mov dx, CRTC_DATA                       ; to zero
     in al, dx
     mov OLD_VERT_TOTAL, al                  ; (save old value first)
     xor al, al
     out dx, al

     mov dx, CRTC_INDEX
     mov al, 07h
     out dx, al                              ; bits 8 & 9 of Vertical Total
     mov dx, CRTC_DATA                       ; are in the OVERFLOW register
     in al, dx                               ; clear them
     mov OLD_OVERFLOW, al
     and al, 11011110b                       ; (save old value first)
     out dx, al

     call END_WRITE                          ; Restore Write prot & AX, DX

     ret
STOP_V_SYNC ENDP

STOP_H_SYNC PROC NEAR
     call INIT_WRITE                         ; Save regs & clear Write protect

     mov dx, CRTC_INDEX
     mov al, 00h                             ; Suppress Horizontal Sync signal
     out dx, al                              ; By setting Horizontal total
     mov dx, CRTC_DATA                       ; to zero
     in al, dx
     mov OLD_HORIZ_TOTAL, al                 ; (save old value first)
     xor al, al
     out dx, al

     call END_WRITE                          ; Restore Write prot & AX, DX 

     ret
STOP_H_SYNC ENDP

RESTART_V_SYNC PROC NEAR
     call INIT_WRITE                         ; This procedure restarts the
                                             ; VSYNC signal by restoring the
     mov dx, CRTC_INDEX                      ; Vertical Total saved by the
     mov al, 06h                             ; STOP_V_SYNC procedure
     out dx, al
     mov dx, CRTC_DATA
     mov al, OLD_VERT_TOTAL
     out dx, al

     mov dx, CRTC_INDEX
     mov al, 07h
     out dx, al
     mov al, OLD_OVERFLOW
     mov dx, CRTC_DATA
     out dx, al

     call END_WRITE

      ret
RESTART_V_SYNC ENDP

RESTART_H_SYNC PROC NEAR
     call INIT_WRITE                         ; This procedure restarts the
                                             ; HSYNC signal by restoring the
     mov dx, CRTC_INDEX                      ; Horizontal Total saved by the
     mov al, 00h                             ; STOP_H_SYNC procedure
     out dx, al
     mov dx, CRTC_DATA
     mov al, OLD_HORIZ_TOTAL
     out dx, al

     call END_WRITE

     ret
RESTART_H_SYNC ENDP

TIMER_HANDLER  PROC FAR
     push ds
     push ax                                 ; Switch to our Data Segment
     mov ax, cs                              ; save old values first
     mov ds, ax

     cmp CURRENT_MODE, POWER_OFF             ; If already in power off mode
     je TH_Return                            ; don't do anything
    
     inc TIMER                               ; increment timer

     push dx
     push cx

     mov ax, TIMER                           ; Convert TIMER from clock ticks
     mov cx, 5                               ; to seconds. 18.2 clk ticks i.e.
     mul cx                                  ; 91/5 clk tick = 1 second
     mov cx, 91
     div cx

     pop cx
     pop dx

     cmp ax, POWER_OFF_TIME                  ; If less than pow off time,
     jb  TH_ChkSuspend                       ; Check for suspen time

     cmp CURRENT_MODE, SUSPEND               ; If already in suspend mode
     je TH_StopHSync                         ; V_SYNC is already stopped
     call STOP_V_SYNC                        ; otherwise stop V_SYNC first

     TH_StopHSync:                           ; Now, stop H_SYNC
     call STOP_H_SYNC
     mov CURRENT_MODE, POWER_OFF
     jmp SHORT TH_Return

     TH_ChkSuspend:                          ; If time < suspend time, do
     cmp ax, SUSPEND_TIME                    ; nothing. Otherwise stop
     jb TH_Return                            ; V_SYNC
     cmp CURRENT_MODE, SUSPEND
     je TH_Return

     call STOP_V_SYNC
     mov CURRENT_MODE, SUSPEND

     TH_Return:
     pop ax                                  ; Restore saved registers and
     pop ds                                  ; Jump to original TIMER vector
     jmp CS:ORIG_08h_VECT
TIMER_HANDLER  ENDP

KEYBOARD_HANDLER PROC FAR
     push ax                                 ; Switch to out Data Segment
     push ds                                 ; save old values first
     mov ax, cs
     mov ds, ax
                                             
     mov TIMER, 0                            ; Key pressed hence reset timer

     cmp CURRENT_MODE, POWER_OFF             ; If in power off mode, restart
     jne KH_ChkSuspend                       ; both H_SYNC and V_SYNC
     call RESTART_H_SYNC                     ; signals
     call RESTART_V_SYNC
     jmp SHORT KH_Return

     KH_ChkSuspend:
     cmp CURRENT_MODE, SUSPEND               ; If in power off mode, restart
     jne KH_Return                           ; V_SYNC signal
     call RESTART_V_SYNC

     KH_Return:
     mov CURRENT_MODE, NORMAL                
     pop ds                                  ; Restore saved registers and
     pop ax                                  ; jump to the original KEYBOARD
     jmp CS:ORIG_09h_VECT                    ; vector
KEYBOARD_HANDLER ENDP

END_RESIDENT LABEL BYTE                      ; Everything beyond this point
                                             ; is useless after initialization
                                             ; Will be discarded

;--------------------------------DATA (TRANSIENT)-----------------------------
INTRODUCTION   db   CR, LF
               db   "DPMS Screen Saver", CR, LF
               db   "    -by Vinay Pai", CR, LF
               db   LF
               db   "A DOS mode screen saver that will switch off DPMS compliant monitors", CR, LF
               db   LF, ENDSTR

HELPSTRING     db   "Any of the following commands may be used: ", CR, LF
               db   LF
               db   "/S=n      - enter suspend mode after 'n' seconds (max: 3600)", CR, LF
               db   "/O=n      - enter power off mode after 'n' seconds (max: 3600)", CR, LF
               db LF
               db   "NOTE: If power off time is less than suspend time, the Screen saver", CR, LF
               db   "      will never enter suspend mode", CR, LF
               db LF, ENDSTR

;--------------------------------CODE (TRANSIENT)-----------------------------
GET_NUMBER PROC NEAR                         ; GET_NUMBER accepts address to
     cmp cx, 0                               ; store decoded number, in DI
     je GN_Return

     GN_NextChar:
     lodsb                                   ; Check if char is a digit
     cmp al, '0'                             
     jb GN_NotDigit
     cmp al, '9'
     ja GN_NotDigit

     push ax                                 ; If so, make it the rightmost
     mov ax, 10                              ; digit in the number being
     mul WORD PTR [di]                       ; decoded
     mov [di], ax
     pop ax
     sub al, '0'
     add [di], al
     loop GN_NextChar
     jmp SHORT GN_Return

     GN_NotDigit:                            ; Return if no more chars
     cmp cx, 0
     je GN_Return

     cmp WORD PTR [di], 0
     jne GN_CleanupAndReturn

     cmp al, ' '                             ; Ignore ' ' and '='
     loope GN_NextChar                       ; unless if occurs in the middle
     cmp al, '='                             ; of a no so that, for example:
     loope GN_NextChar                       ; 12 34 is read as 12 not 1234

     GN_CleanupAndReturn:                    ; We have read one character too
     dec si                                  ; many. Push it back, so that
     inc cx                                  ; something like /s=10/o=10 is
                                             ; decoded correctly
     GN_Return:
     ret
GET_NUMBER ENDP

PARSE_COMMAND_LINE PROC NEAR
     cmp BYTE PTR DS:CMD_LEN, 0              ; Check for valid cmdline
     jne PC_CmdlineOK
     ret                                     ; Return in cmdline length=0

     PC_CmdLineOK:
     cld
     xor ch, ch
     mov cl, DS:CMD_LEN
     mov si, CMD_LINE

     PC_NextChar:
     lodsb
     cmp al, '/'                             ; Check for '/' or '-'
     je PC_Analyze
     cmp al, '-'
     je PC_Analyze
     loop PC_NextChar
     PC_EndScan:
     ret

     PC_Analyze:                             ; Identify the command line
     cmp cx, 0                               ; switch specified
     je PC_EndScan
     lodsb
     cmp al, '?'
     je PC_Help
     cmp al, 'H'
     je PC_Help                              ; and branch to the required
     cmp al, 'S'                             ; routine
     je PC_Suspend
     cmp al, 'O'
     je PC_PowerOff
     cmp al, 'h'
     je PC_Help                              
     cmp al, 's'                             
     je PC_Suspend
     cmp al, 'o'
     je PC_PowerOff

     inc cx
     dec si
     loop PC_NextChar                        ; continue processing the cmdline
     jmp SHORT PC_EndParse

     PC_Help:
     mov dx, offset HELPSTRING               ; Command line help requested
     mov ah, 9                               ; display help string
     int 21h                                 ; and quit without remaining
     mov ax, 4C00h                           ; resident
     int 21h

     PC_Suspend:                             ; /S=n is specified
     mov SUSPEND_TIME, 0                     ; find and set the value of
     mov di, OFFSET SUSPEND_TIME             ; SUSPEND_TIME
     call GET_NUMBER                         
     cmp SUSPEND_TIME, 0
     je PC_Help
     cmp SUSPEND_TIME, 3600
     ja PC_Help
     cmp cx, 0
     je PC_EndParse
     loop PC_NextChar

     PC_PowerOff:                            ; /O=n is specified
     mov POWER_OFF_TIME, 0                   ; find and set the value of
     mov di, OFFSET POWER_OFF_TIME           ; POWER_OFF_TIME
     call GET_NUMBER
     cmp POWER_OFF_TIME, 0
     je PC_Help
     cmp POWER_OFF_TIME, 3600
     ja PC_Help
     loop PC_NextChar

     PC_EndParse:
     ret
PARSE_COMMAND_LINE ENDP

MAIN PROC NEAR
     mov dx, offset INTRODUCTION             ; display the startup
     mov ah, 9                               ; message
     int 21h

     mov ax, 3508h                           ; save the old TIMER (IRQ0)
     int 21h                                 ; vector
     mov word ptr ORIG_08h_VECT, bx
     mov word ptr ORIG_08h_VECT + 2, es

     mov ax, 3509h                           ; save the old KEYBOARD (IRQ1)
     int 21h                                 ; vector                    
     mov WORD PTR ORIG_09h_VECT, bx
     mov WORD PTR ORIG_09h_VECT + 2, es

     call PARSE_COMMAND_LINE                 ; Check the command line params

     mov ax, 2508h                           ; Revector the TIMER interrupt 
     mov dx, OFFSET TIMER_HANDLER            ; to our own ISR
     int 21h

     mov ax, 2509h                           ; Revector the KEYBOARD interrupt 
     mov dx, OFFSET KEYBOARD_HANDLER         ; to our own ISR
     int 21h

     mov ax, DS:2Ch                          ; Release the local environment
     mov es, ax                              ; copy to reduce the TSR size
     mov ah, 49h
     int 21h

     mov dx, OFFSET END_RESIDENT             ; Calculate TSR size
     add dx, 0Eh                             ; and stay resident
     shr dx, 4
     mov ax, 3100h
     int 21h
MAIN ENDP

CSEG ENDS
     END ENTPT
     
